/* Demonstration program showing use of the Array package and         */
/* Complex class.                                                     */
/*                                                                    */
/* CMATMUL: This program computes the matrix product C = A x B        */
/* where A, B, and C are matrices of complex entries.                 */
/*                                                                    */
/* Results are checked for validity.                                  */

import com.ibm.math.Complex;	// import the Complex class
import com.ibm.math.array.*;	// import the Array package

public class cmatmul {

  private final static int min(int a, int b) {
    return (a<b) ? a : b;
  }

  public static void main(String[] args) {

    int n = 100;	// problem size

    /*
     * Create matrices A, B, and C.
     */
    ComplexArray2D a = new ComplexArray2D(n,n);
    ComplexArray2D b = new ComplexArray2D(n,n);
    ComplexArray2D c = new ComplexArray2D(n,n);

    int i;
    int j;
    double t1;
    double t2;
    double time;
    double ops;
    double mflops;

    /*
     * Initialize the data.
     */
    for( j = 0; j <= (n - 1); j++) {
      for( i = 0; i <= (n - 1); i++) {
        a.set(j,i,(new Complex(1.0,1.0)));
        b.set(j,i,(new Complex(1.0,1.0)));
        c.set(j,i,(new Complex(0.0,0.0)));
      };
    };

    /*
     * Do the computation.
     */
    t1 = (1e-3*System.currentTimeMillis());
    cmm(a,b,c,n);
    t2 = (1e-3*System.currentTimeMillis());

    /*
     * Print performance results.
     */
    time = (t2 - t1);
    ops = (((8.0 * n) * n) * n);
    mflops = ((1.0e-6 * ops) / time);
    System.out.print("Problem size: "+n+"x"+n);
    System.out.println(" Time: "+time+" Mflops: "+mflops);

    /*
     * Check result for correctness.
     */
    validate(c,n);
  }

  public static void validate(ComplexArray2D c,int n) {
    int i;
    int j;
    int result;
    double eps;
    result = 1;
    eps = 1.0e-9;
    for( j = 0; j <= (n - 1); j++) {
      for( i = 0; i <= (n - 1); i++) {
        if (((Complex.abs((c.get(j,i).minus((new Complex(0.0,(2.0 * n))))))) > eps)) {
          result = 0;
        };
      };
    };

    if (result == 1) {
      System.out.println("Result is correct!");
    } else {
      System.out.println("Result is incorrect!");
    }
  }

  public static void cmm(ComplexArray2D a,ComplexArray2D b,ComplexArray2D c,int n) {
    int i;
    int io;
    int j;
    int jo;
    int k;
    int ko;
    int bs = 88;
    Complex s0 = new Complex(0.0,0.0);
    Complex s1 = new Complex(0.0,0.0);
    Complex s2 = new Complex(0.0,0.0);
    Complex s3 = new Complex(0.0,0.0);
    Complex s4 = new Complex(0.0,0.0);
    Complex s5 = new Complex(0.0,0.0);
    Complex s6 = new Complex(0.0,0.0);
    Complex s7 = new Complex(0.0,0.0);
    for( io = 0; io <= (n - 1); io += bs) {
      for( jo = 0; jo <= (n - 1); jo += bs) {
        for( ko = 0; ko <= (n - 1); ko += bs) {
          for( i = io; i <= ((min(n,(io + bs))) - 4); i += 4) {
            for( j = jo; j <= ((min(n,(jo + bs))) - 2); j += 2) {
              s0.assign(c.get(j,i));
              s1.assign(c.get((j + 1),i));
              s2.assign(c.get(j,(i + 1)));
              s3.assign(c.get((j + 1),(i + 1)));
              s4.assign(c.get(j,(i + 2)));
              s5.assign(c.get((j + 1),(i + 2)));
              s6.assign(c.get(j,(i + 3)));
              s7.assign(c.get((j + 1),(i + 3)));
              for( k = ko; k <= ((min(n,(ko + bs))) - 1); k++) {
                s0.assign((s0.plus((a.get(k,i).times(b.get(j,k))))));
                s1.assign((s1.plus((a.get(k,i).times(b.get((j + 1),k))))));
                s2.assign((s2.plus((a.get(k,(i + 1)).times(b.get(j,k))))));
                s3.assign((s3.plus((a.get(k,(i + 1)).times(b.get((j + 1),k))))));
                s4.assign((s4.plus((a.get(k,(i + 2)).times(b.get(j,k))))));
                s5.assign((s5.plus((a.get(k,(i + 2)).times(b.get((j + 1),k))))));
                s6.assign((s6.plus((a.get(k,(i + 3)).times(b.get(j,k))))));
                s7.assign((s7.plus((a.get(k,(i + 3)).times(b.get((j + 1),k))))));
              };
              c.set(j,i,s0);
              c.set((j + 1),i,s1);
              c.set(j,(i + 1),s2);
              c.set((j + 1),(i + 1),s3);
              c.set(j,(i + 2),s4);
              c.set((j + 1),(i + 2),s5);
              c.set(j,(i + 3),s6);
              c.set((j + 1),(i + 3),s7);
            };
            for( j = j; j <= ((min(n,(jo + bs))) - 1); j++) {
              s0.assign(c.get(j,i));
              s1.assign(c.get(j,(i + 1)));
              s2.assign(c.get(j,(i + 2)));
              s3.assign(c.get(j,(i + 3)));
              for( k = ko; k <= ((min(n,(ko + bs))) - 1); k++) {
                s0.assign((s0.plus((a.get(k,i).times(b.get(j,k))))));
                s1.assign((s1.plus((a.get(k,(i + 1)).times(b.get(j,k))))));
                s2.assign((s2.plus((a.get(k,(i + 2)).times(b.get(j,k))))));
                s3.assign((s3.plus((a.get(k,(i + 3)).times(b.get(j,k))))));
              };
              c.set(j,i,s0);
              c.set(j,(i + 1),s1);
              c.set(j,(i + 2),s2);
              c.set(j,(i + 3),s3);
            };
          };
          for( i = i; i <= ((min(n,(io + bs))) - 1); i++) {
            for( j = jo; j <= ((min(n,(jo + bs))) - 1); j++) {
              s0.assign(c.get(j,i));
              for( k = ko; k <= ((min(n,(ko + bs))) - 1); k++) {
                s0.assign((s0.plus((a.get(k,i).times(b.get(j,k))))));
              };
              c.set(j,i,s0);
            };
          };
        };
      };
    };
  }
}
